/*
 * Cifra RSA
 *
 * Criptoanlise da cifra RSA
 */

#include <cstdio> // para poder usar FILE
#include <cmath>  // sqrt

#ifndef CIFRARSA
#define CIFRARSA

class CifraRSA {
private:

    int comprimentoVector = 5000;
    /*
	 * Verifica se um dado natural  primo
	 * --> cand: natural
	 * <-- p: p=true, n  primo, p=false, no  primo
     */
	bool testaPrimo(long int);
public:
	/*
	 * Crivo de Eratostenes
	 * --> n: inteiro
	 * <-- primos: vector de inteiros, contendo os primos at n (inclusiv)
	 */
	void crivoEratostenes(long int n, long int *contaPrimos, long int primos[]); 
	/* 
	 * Mtodo das Divises --- Criptoanlise da cifra RSA
	 * Utiliza o Crivo de Eratostenes para obter uma lista de primos a testar
	 * --> n: inteiro, produto de dois primos
	 * <-- p,q: inteiros, primos tais que pq = n
	 */
	void metodoDivisoes(long int n, long int *p, long int *q);	
	/* 
	 * Mtodo da frmula geradora de primos 6k+-1 --- Criptoanlise da cifra RSA
	 * --> n: inteiro, produto de dois primos
	 * <-- p,q: inteiros, primos tais que pq = n
	 */
	void formulaGeradora(long int n, long int *p, long int *q);
	
	// Mtodo de Fermat
	void metodoFermat(long int n, long int *p, long int *q);
	
	/*
	 *  Gerar das Chaves
	 * --> p, q: primos (de grande dimenso)
	 * <-- (e,n),(d,n): n=p*q, (e,n) - chave pblica, (d,n) - chave privada
	 */
	void gerarChavesRSA(long int p, long int q, long int *n, long int *e, long int *d);
	
	/*
	 * Gerar um primo (de grande dimenso) --- crivo de Eratostenes
	 */
	 
	/* cifrar RSA
	 * --> m: inteiro (de grande dimenso), bloco da mensagem
	 *     (e,n): chave pblica
	 * <-- c: inteiro (de grande dimenso) resultado da encriptao de m, c=cifrarRSA(m,e,)
	 */
	long int cifrarRSA(long int m, long int e, long int n); 
	
	/* decifrar RSA 
	 * --> c: inteiro (de grande dimenso), bloco da mensagem encriptada
	 *     (d,n): chave privada
	 * <-- m: inteiro (de grande dimenso) resultado da desencriptao de m=decifrarRSA(c,d,n)
	 */
	long int decifrarRSA(long int m, long int e, long int n); 
	
};

#endif
